@@ -37,7 +37,6 @@
 #include "error.h"
 #include "lchown.h"
 #include "quote.h"
-#include "savedir.h"
 #include "chown-core.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
@@ -56,7 +55,6 @@ struct group *getgrgid ();
 #endif
 
 char *parse_user_spec ();
-void strip_trailing_slashes ();
 
 /* The name the program was run with. */
 char *program_name;
@@ -104,35 +102,41 @@ Usage: %s [OPTION]... OWNER[:[GROUP]] FILE...\n\
   or:  %s [OPTION]... --reference=RFILE FILE...\n\
 "),
 	      program_name, program_name, program_name);
-      printf (_("\
+      fputs (_("\
 Change the owner and/or group of each FILE to OWNER and/or GROUP.\n\
 \n\
   -c, --changes          like verbose but report only when a change is made\n\
       --dereference      affect the referent of each symbolic link, rather\n\
                          than the symbolic link itself\n\
+"), stdout);
+      fputs (_("\
   -h, --no-dereference   affect symbolic links instead of any referenced file\n\
                          (available only on systems that can change the\n\
                          ownership of a symlink)\n\
+"), stdout);
+      fputs (_("\
       --from=CURRENT_OWNER:CURRENT_GROUP\n\
                          change the owner and/or group of each file only if\n\
                          its current owner and/or group match those specified\n\
                          here.  Either may be omitted, in which case a match\n\
                          is not required for the omitted attribute.\n\
+"), stdout);
+      fputs (_("\
   -f, --silent, --quiet  suppress most error messages\n\
       --reference=RFILE  use RFILE's owner and group rather than\n\
                          the specified OWNER:GROUP values\n\
   -R, --recursive        operate on files and directories recursively\n\
   -v, --verbose          output a diagnostic for every file processed\n\
-      --help             display this help and exit\n\
-      --version          output version information and exit\n\
+"), stdout);
+      fputs (HELP_OPTION_DESCRIPTION, stdout);
+      fputs (VERSION_OPTION_DESCRIPTION, stdout);
+      fputs (_("\
 \n\
-"));
-      printf (_("\
 Owner is unchanged if missing.  Group is unchanged if missing, but changed\n\
 to login group if implied by a `:'.  OWNER and GROUP may be numeric as well\n\
 as symbolic.\n\
-"));
-      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+"), stdout);
+      printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
   exit (status);
 }
@@ -173,11 +177,11 @@ main (int argc, char **argv)
 	case FROM_OPTION:
 	  {
 	    char *u_dummy, *g_dummy;
-	    const char *e = parse_user_spec (argv[optind],
+	    const char *e = parse_user_spec (optarg,
 					     &old_uid, &old_gid,
 					     &u_dummy, &g_dummy);
 	    if (e)
-	      error (1, 0, "%s: %s", quote (argv[optind]), e);
+	      error (1, 0, "%s: %s", quote (optarg), e);
 	    break;
 	  }
 	case 'R':
@@ -213,7 +217,8 @@ main (int argc, char **argv)
       struct stat ref_stats;
 
       if (stat (reference_file, &ref_stats))
-	error (1, errno, _("getting attributes of %s"), quote (reference_file));
+	error (1, errno, _("failed to get attributes of %s"),
+	       quote (reference_file));
 
       uid = ref_stats.st_uid;
       gid = ref_stats.st_gid;
@@ -235,11 +240,8 @@ main (int argc, char **argv)
     }
 
   for (; optind < argc; ++optind)
-    {
-      strip_trailing_slashes (argv[optind]);
-      errors |= change_file_owner (1, argv[optind], uid, gid,
-				   old_uid, old_gid, &chopt);
-    }
+    errors |= change_file_owner (1, argv[optind], uid, gid,
+				 old_uid, old_gid, &chopt);
 
   chopt_free (&chopt);
 
