@@ -1,5 +1,5 @@
 /* copy.c -- core functions for copying files and directories
-   Copyright (C) 89, 90, 91, 1995-2001 Free Software Foundation.
+   Copyright (C) 89, 90, 91, 1995-2002 Free Software Foundation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -26,16 +26,25 @@
 #include <assert.h>
 #include <sys/types.h>
 
+#if HAVE_HURD_H
+# include <hurd.h>
+#endif
+
 #include "system.h"
 #include "error.h"
 #include "backupfile.h"
 #include "savedir.h"
 #include "copy.h"
 #include "cp-hash.h"
+#include "hash.h"
+#include "hash-pjw.h"
+#include "same.h"
 #include "dirname.h"
+#include "full-write.h"
 #include "path-concat.h"
 #include "quote.h"
 #include "same.h"
+#include "xreadlink.h"
 
 #define DO_CHOWN(Chown, File, New_uid, New_gid)				\
   (Chown (File, New_uid, New_gid)					\
@@ -48,6 +57,11 @@
 #define SAME_GROUP(A, B) ((A).st_gid == (B).st_gid)
 #define SAME_OWNER_AND_GROUP(A, B) (SAME_OWNER (A, B) && SAME_GROUP (A, B))
 
+#define UNWRITABLE(File_name, File_mode)		\
+  ( /* euidaccess is not meaningful for symlinks */	\
+    ! S_ISLNK (File_mode)				\
+    && euidaccess (File_name, W_OK) != 0)
+
 struct dir_list
 {
   struct dir_list *parent;
@@ -55,7 +69,17 @@ struct dir_list
   dev_t dev;
 };
 
-int full_write ();
+/* Describe a just-created or just-renamed destination file.  */
+struct F_triple
+{
+  char const* name;
+  ino_t st_ino;
+  dev_t st_dev;
+};
+
+/* Initial size of the above hash table.  */
+#define DEST_INFO_INITIAL_CAPACITY 61
+
 int euidaccess ();
 int yesno ();
 
@@ -86,9 +110,11 @@ get_dest_mode (const struct cp_options *option, mode_t mode)
   if (option->set_mode)
     return option->mode;
 
-  /* Honor the umask for `cp', but not for `mv' or `cp -p'.  */
-  if (!option->move_mode && !option->preserve_chmod_bits)
-    mode &= option->umask_kill;
+  /* Honor the umask for `cp', but not for `mv' or `cp -p'.
+     In addition, `cp' without -p must clear the set-user-ID and set-group-ID
+     bits.  POSIX requires it do that when creating new files.  */
+  if (!option->move_mode && !option->preserve_mode)
+    mode &= (option->umask_kill & ~(S_ISUID | S_ISGID));
 
   return mode;
 }
@@ -128,7 +154,7 @@ copy_dir (const char *src_path_in, const char *dst_path_in, int new_dst,
   struct cp_options non_command_line_options = *x;
   int ret = 0;
 
-  name_space = savedir (src_path_in, src_sb->st_size);
+  name_space = savedir (src_path_in);
   if (name_space == NULL)
     {
       /* This diagnostic is a bit vague because savedir can fail in
@@ -157,10 +183,7 @@ copy_dir (const char *src_path_in, const char *dst_path_in, int new_dst,
 			    &local_copy_into_self, NULL);
       *copy_into_self |= local_copy_into_self;
 
-      /* Free the memory for `src_path'.  The memory for `dst_path'
-	 cannot be deallocated, since it is used to create multiple
-	 hard links.  */
-
+      free (dst_path);
       free (src_path);
 
       namep += strlen (namep) + 1;
@@ -176,17 +199,20 @@ copy_dir (const char *src_path_in, const char *dst_path_in, int new_dst,
    Use DST_MODE as the 3rd argument in the call to open.
    X provides many option settings.
    Return 0 if successful, -1 if an error occurred.
-   *NEW_DST is as in copy_internal.  */
+   *NEW_DST is as in copy_internal.  SRC_SB is the result
+   of calling xstat (aka stat in this case) on SRC_PATH.  */
 
 static int
 copy_reg (const char *src_path, const char *dst_path,
-	  const struct cp_options *x, mode_t dst_mode, int *new_dst)
+	  const struct cp_options *x, mode_t dst_mode, int *new_dst,
+	  struct stat const *src_sb)
 {
   char *buf;
   int buf_size;
   int dest_desc;
   int source_desc;
   struct stat sb;
+  struct stat src_open_sb;
   char *cp;
   int *ip;
   int return_val = 0;
@@ -197,19 +223,28 @@ copy_reg (const char *src_path, const char *dst_path,
   source_desc = open (src_path, O_RDONLY);
   if (source_desc < 0)
     {
-      /* If SRC_PATH doesn't exist, then chances are good that the
-	 user did something like this `cp --backup foo foo': and foo
-	 existed to start with, but copy_internal renamed DST_PATH
-	 with the backup suffix, thus also renaming SRC_PATH.  */
-      if (errno == ENOENT)
-	error (0, 0, _("%s and %s are the same file"),
-	       quote_n (0, src_path), quote_n (1, dst_path));
-      else
-	error (0, errno, _("cannot open %s for reading"), quote (src_path));
-
+      error (0, errno, _("cannot open %s for reading"), quote (src_path));
       return -1;
     }
 
+  if (fstat (source_desc, &src_open_sb))
+    {
+      error (0, errno, _("cannot fstat %s"), quote (src_path));
+      return_val = -1;
+      goto close_src_desc;
+    }
+
+  /* Compare the source dev/ino from the open file to the incoming,
+     saved ones obtained via a previous call to stat.  */
+  if (! SAME_INODE (*src_sb, src_open_sb))
+    {
+      error (0, 0,
+	     _("skipping file %s, as it was replaced while being copied"),
+	     quote (src_path));
+      return_val = -1;
+      goto close_src_desc;
+    }
+
   /* These semantics are required for cp.
      The if-block will be taken in move_mode.  */
   if (*new_dst)
@@ -244,7 +279,7 @@ copy_reg (const char *src_path, const char *dst_path,
       goto close_src_desc;
     }
 
-  /* Find out the optimal buffer size.  */
+  /* Determine the optimal buffer size.  */
 
   if (fstat (dest_desc, &sb))
     {
@@ -283,7 +318,7 @@ copy_reg (const char *src_path, const char *dst_path,
 
   for (;;)
     {
-      int n_read = read (source_desc, buf, buf_size);
+      ssize_t n_read = read (source_desc, buf, buf_size);
       if (n_read < 0)
 	{
 #ifdef EINTR
@@ -336,7 +371,8 @@ copy_reg (const char *src_path, const char *dst_path,
 	}
       if (ip == 0)
 	{
-	  if (full_write (dest_desc, buf, n_read) < 0)
+	  size_t n = n_read;
+	  if (full_write (dest_desc, buf, n) != n)
 	    {
 	      error (0, errno, _("writing %s"), quote (dst_path));
 	      return_val = -1;
@@ -354,12 +390,12 @@ copy_reg (const char *src_path, const char *dst_path,
     {
 #if HAVE_FTRUNCATE
       /* Write a null character and truncate it again.  */
-      if (full_write (dest_desc, "", 1) < 0
+      if (full_write (dest_desc, "", 1) != 1
 	  || ftruncate (dest_desc, n_read_total) < 0)
 #else
       /* Seek backwards one character and write a null.  */
       if (lseek (dest_desc, (off_t) -1, SEEK_CUR) < 0L
-	  || full_write (dest_desc, "", 1) < 0)
+	  || full_write (dest_desc, "", 1) != 1)
 #endif
 	{
 	  error (0, errno, _("writing %s"), quote (dst_path));
@@ -453,15 +489,39 @@ same_file_ok (const char *src_path, const struct stat *src_sb,
 	return 1;
     }
 
-  /* The backup code ensures there's a copy, so it's ok to remove
-     any destination file.  But there's one exception: when both
+  /* The backup code ensures there's a copy, so it's usually ok to
+     remove any destination file.  One exception is when both
      source and destination are the same directory entry.  In that
      case, moving the destination file aside (in making the backup)
      would also rename the source file and result in an error.  */
   if (x->backup_type != none)
     {
       if (!same_link)
-	return 1;
+	{
+	  /* In copy mode when dereferencing symlinks, if the source is a
+	     symlink and the dest is not, then backing up the destination
+	     (moving it aside) would make it a dangling symlink, and the
+	     subsequent attempt to open it in copy_reg would fail with
+	     a misleading diagnostic.  Avoid that by returning zero in
+	     that case so the caller can make cp (or mv when it has to
+	     resort to reading the source file) fail now.  */
+
+	  /* FIXME-note: even with the following kludge, we can still provoke
+	     the offending diagnostic.  It's just a little harder to do :-)
+	     $ rm -f a b c; touch c; ln -s c b; ln -s b a; cp -b a b
+	     cp: cannot open `a' for reading: No such file or directory
+	     That's misleading, since a subsequent `ls' shows that `a'
+	     is still there.
+	     One solution would be to open the source file *before* moving
+	     aside the destination, but that'd involve a big rewrite. */
+	  if ( ! x->move_mode
+	       && x->dereference != DEREF_NEVER
+	       && S_ISLNK (src_sb_link->st_mode)
+	       && ! S_ISLNK (dst_sb_link->st_mode))
+	    return 0;
+
+	  return 1;
+	}
 
       return ! same_name (src_path, dst_path);
     }
@@ -541,13 +601,180 @@ same_file_ok (const char *src_path, const struct stat *src_sb,
   return 0;
 }
 
+static void
+overwrite_prompt (char const *dst_path, struct stat const *dst_sb)
+{
+  if (euidaccess (dst_path, W_OK) != 0)
+    {
+      fprintf (stderr,
+	       _("%s: overwrite %s, overriding mode %04lo? "),
+	       program_name, quote (dst_path),
+	       (unsigned long) (dst_sb->st_mode & CHMOD_MODE_BITS));
+    }
+  else
+    {
+      fprintf (stderr, _("%s: overwrite %s? "),
+	       program_name, quote (dst_path));
+    }
+}
+
+/* Hash an F_triple.  */
+static unsigned int
+triple_hash (void const *x, unsigned int table_size)
+{
+  struct F_triple const *p = x;
+
+  /* Also take the name into account, so that when moving N hard links to the
+     same file (all listed on the command line) all into the same directory,
+     we don't experience any N^2 behavior.  */
+  /* FIXME-maybe: is it worth the overhead of doing this
+     just to avoid N^2 in such an unusual case?  N would have
+     to be very large to make the N^2 factor noticable, and
+     one would probably encounter a limit on the length of
+     a command line before it became a problem.  */
+  unsigned int tmp = hash_pjw (p->name, table_size);
+
+  /* Ignoring the device number here should be fine.  */
+  return (tmp | p->st_ino) % table_size;
+}
+
+/* Hash an F_triple.  */
+static unsigned int
+triple_hash_no_name (void const *x, unsigned int table_size)
+{
+  struct F_triple const *p = x;
+
+  /* Ignoring the device number here should be fine.  */
+  return p->st_ino % table_size;
+}
+
+/* Compare two F_triple structs.  */
+static bool
+triple_compare (void const *x, void const *y)
+{
+  struct F_triple const *a = x;
+  struct F_triple const *b = y;
+  return (SAME_INODE (*a, *b) && same_name (a->name, b->name)) ? true : false;
+}
+
+/* Free an F_triple.  */
+static void
+triple_free (void *x)
+{
+  struct F_triple *a = x;
+  free ((char *) (a->name));
+  free (a);
+}
+
+/* Initialize the hash table implementing a set of F_triple entries
+   corresponding to destination files.  */
+void
+dest_info_init (struct cp_options *x)
+{
+  x->dest_info
+    = hash_initialize (DEST_INFO_INITIAL_CAPACITY,
+		       NULL,
+		       triple_hash,
+		       triple_compare,
+		       triple_free);
+}
+
+/* Initialize the hash table implementing a set of F_triple entries
+   corresponding to source files listed on the command line.  */
+void
+src_info_init (struct cp_options *x)
+{
+
+  /* Note that we use triple_hash_no_name here.
+     Contrast with the use of triple_hash above.
+     That is necessary because a source file may be specified
+     in many different ways.  We want to warn about this
+       cp a a d/
+     as well as this:
+       cp a ./a d/
+  */
+  x->src_info
+    = hash_initialize (DEST_INFO_INITIAL_CAPACITY,
+		       NULL,
+		       triple_hash_no_name,
+		       triple_compare,
+		       triple_free);
+}
+
+/* Return nonzero if there is an entry in hash table, HT,
+   for the file described by FILENAME and STATS.
+   Otherwise, return zero.  */
+static int
+seen_file (Hash_table const *ht, char const *filename,
+	   struct stat const *stats)
+{
+  struct F_triple new_ent;
+
+  if (ht == NULL)
+    return 0;
+
+  new_ent.name = filename;
+  new_ent.st_ino = stats->st_ino;
+  new_ent.st_dev = stats->st_dev;
+
+  return !!hash_lookup (ht, &new_ent);
+}
+
+/* Record destination filename, FILENAME, and dev/ino from *STATS,
+   in the hash table, HT.  If HT is NULL, return immediately.
+   If STATS is NULL, call lstat on FILENAME to get the device
+   and inode numbers.  If that lstat fails, simply return.
+   If memory allocation fails, exit immediately.  */
+static void
+record_file (Hash_table *ht, char const *filename,
+	     struct stat const *stats)
+{
+  struct F_triple *ent;
+
+  if (ht == NULL)
+    return;
+
+  ent = (struct F_triple *) xmalloc (sizeof *ent);
+  ent->name = xstrdup (filename);
+  if (stats)
+    {
+      ent->st_ino = stats->st_ino;
+      ent->st_dev = stats->st_dev;
+    }
+  else
+    {
+      struct stat sb;
+      if (lstat (filename, &sb))
+	return;
+      ent->st_ino = sb.st_ino;
+      ent->st_dev = sb.st_dev;
+    }
+
+  {
+    struct F_triple *ent_from_table = hash_insert (ht, ent);
+    if (ent_from_table == NULL)
+      {
+	/* Insertion failed due to lack of memory.  */
+	xalloc_die ();
+      }
+
+    if (ent_from_table != ent)
+      {
+	/* There was alread a matching entry in the table, so ENT was
+	   not inserted.  Free it.  */
+	triple_free (ent);
+      }
+  }
+}
+
 /* Copy the file SRC_PATH to the file DST_PATH.  The files may be of
    any type.  NEW_DST should be nonzero if the file DST_PATH cannot
    exist because its parent directory was just created; NEW_DST should
    be zero if DST_PATH might already exist.  DEVICE is the device
    number of the parent directory, or 0 if the parent of this file is
    not known.  ANCESTORS points to a linked, null terminated list of
-   devices and inodes of parent directories of SRC_PATH.
+   devices and inodes of parent directories of SRC_PATH.  COMMAND_LINE_ARG
+   is nonzero iff SRC_PATH was specified on the command line.
    Set *COPY_INTO_SELF to nonzero if SRC_PATH is a parent of (or the
    same as) DST_PATH;  otherwise, set it to zero.
    Return 0 if successful, 1 if an error occurs. */
@@ -558,7 +785,7 @@ copy_internal (const char *src_path, const char *dst_path,
 	       dev_t device,
 	       struct dir_list *ancestors,
 	       const struct cp_options *x,
-	       int move_mode,
+	       int command_line_arg,
 	       int *copy_into_self,
 	       int *rename_succeeded)
 {
@@ -566,13 +793,20 @@ copy_internal (const char *src_path, const char *dst_path,
   struct stat dst_sb;
   mode_t src_mode;
   mode_t src_type;
-  char *earlier_file;
+  char *earlier_file = NULL;
   char *dst_backup = NULL;
   int backup_succeeded = 0;
-  int rename_errno;
   int delayed_fail;
   int copied_as_regular = 0;
   int ran_chown = 0;
+  int preserve_metadata;
+
+  /* move_mode is set to the value from the `options' parameter for the
+     first copy_internal call.  For any subsequent recursive call, it must
+     be zero.  This is because if we're moving (via mv) a hierarchy and
+     end up having to recurse, it means the initial rename failed and so we
+     are in the process of *copy*ing all of the parts, not renaming them.  */
+  int move_mode = (command_line_arg ? x->move_mode : 0);
 
   if (move_mode && rename_succeeded)
     *rename_succeeded = 0;
@@ -584,18 +818,9 @@ copy_internal (const char *src_path, const char *dst_path,
       return 1;
     }
 
-  /* We wouldn't insert a node unless nlink > 1, except that we need to
-     find created files so as to not copy infinitely if a directory is
-     copied into itself.  */
-
-  /* Associate the destination path with the source device and inode
-     so that if we encounter a matching dev/ino pair in the source tree
-     we can arrange to create a hard link between the corresponding names
-     in the destination tree.  */
-  earlier_file = remember_copied (dst_path, src_sb.st_ino, src_sb.st_dev);
+  src_type = src_sb.st_mode;
 
   src_mode = src_sb.st_mode;
-  src_type = src_sb.st_mode;
 
   if (S_ISDIR (src_type) && !x->recursive)
     {
@@ -603,6 +828,24 @@ copy_internal (const char *src_path, const char *dst_path,
       return 1;
     }
 
+  /* Detect the case in which the same source file appears more than
+     once on the command line and no backup option has been selected.
+     If so, simply warn and don't copy it the second time.
+     This check is enabled only if x->src_info is non-NULL.  */
+  if (command_line_arg)
+    {
+      if ( ! S_ISDIR (src_sb.st_mode)
+	   && x->backup_type == none
+	   && seen_file (x->src_info, src_path, &src_sb))
+	{
+	  error (0, 0, _("warning: source file %s specified more than once"),
+		 quote (src_path));
+	  return 0;
+	}
+
+      record_file (x->src_info, src_path, &src_sb);
+    }
+
   if (!new_dst)
     {
       if ((*(x->xstat)) (dst_path, &dst_sb))
@@ -632,12 +875,32 @@ copy_internal (const char *src_path, const char *dst_path,
 	      return 1;
 	    }
 
-	  if (S_ISDIR (src_type) && !S_ISDIR (dst_sb.st_mode))
+	  if (!S_ISDIR (dst_sb.st_mode))
 	    {
-	      error (0, 0,
+	      if (S_ISDIR (src_type))
+		{
+		  error (0, 0,
 		     _("cannot overwrite non-directory %s with directory %s"),
-		     quote_n (0, dst_path), quote_n (1, src_path));
-	      return 1;
+			 quote_n (0, dst_path), quote_n (1, src_path));
+		  return 1;
+		}
+
+	      /* Don't let the user destroy their data, even if they try hard:
+		 This mv command must fail (likewise for cp):
+		   rm -rf a b c; mkdir a b c; touch a/f b/f; mv a/f b/f c
+		 Otherwise, the contents of b/f would be lost.
+		 In the case of `cp', b/f would be lost if the user simulated
+		 a move using cp and rm.
+		 Note that it works fine if you use --backup=numbered.  */
+	      if (command_line_arg
+		  && x->backup_type != numbered
+		  && seen_file (x->dest_info, dst_path, &dst_sb))
+		{
+		  error (0, 0,
+			 _("will not overwrite just-created %s with %s"),
+			 quote_n (0, dst_path), quote_n (1, src_path));
+		  return 1;
+		}
 	    }
 
 	  if (!S_ISDIR (src_type))
@@ -651,25 +914,54 @@ copy_internal (const char *src_path, const char *dst_path,
 		}
 
 	      if (x->update && MTIME_CMP (src_sb, dst_sb) <= 0)
-		return 0;
+		{
+		  /* We're using --update and the source file is older
+		     than the destination file, so there is no need to
+		     copy or move.  */
+		  /* Pretend the rename succeeded, so the caller (mv)
+		     doesn't end up removing the source file.  */
+		  if (rename_succeeded)
+		    *rename_succeeded = 1;
+		  return 0;
+		}
 	    }
 
-	  if (!S_ISDIR (src_type) && x->interactive)
+	  /* When there is an existing destination file, we may end up
+	     returning early, and hence not copying/moving the file.
+	     This may be due to an interactive `negative' reply to the
+	     prompt about the existing file.  It may also be due to the
+	     use of the --reply=no option.  */
+	  if (!S_ISDIR (src_type))
 	    {
-	      if (euidaccess (dst_path, W_OK) != 0)
+	      /* cp and mv treat -i and -f differently.  */
+	      if (x->move_mode)
 		{
-		  fprintf (stderr,
-			   _("%s: overwrite %s, overriding mode %04lo? "),
-			   program_name, quote (dst_path),
-			   (unsigned long) (dst_sb.st_mode & CHMOD_MODE_BITS));
+		  if ((x->interactive == I_ALWAYS_NO
+		       && UNWRITABLE (dst_path, dst_sb.st_mode))
+		      || ((x->interactive == I_ASK_USER
+			   || (x->interactive == I_UNSPECIFIED
+			       && x->stdin_tty
+			       && UNWRITABLE (dst_path, dst_sb.st_mode)))
+			  && (overwrite_prompt (dst_path, &dst_sb), 1)
+			  && ! yesno ()))
+		    {
+		      /* Pretend the rename succeeded, so the caller (mv)
+			 doesn't end up removing the source file.  */
+		      if (rename_succeeded)
+			*rename_succeeded = 1;
+		      return 0;
+		    }
 		}
 	      else
 		{
-		  fprintf (stderr, _("%s: overwrite %s? "),
-			   program_name, quote (dst_path));
+		  if (x->interactive == I_ALWAYS_NO
+		      || (x->interactive == I_ASK_USER
+			  && (overwrite_prompt (dst_path, &dst_sb), 1)
+			  && ! yesno ()))
+		    {
+		      return 0;
+		    }
 		}
-	      if (!yesno ())
-		return (move_mode ? 1 : 0);
 	    }
 
 	  if (move_mode)
@@ -701,12 +993,14 @@ copy_internal (const char *src_path, const char *dst_path,
 
 	      /* Detect (and fail) when creating the backup file would
 		 destroy the source file.  Before, running the commands
-		 cd /tmp; rm -f a a~; : > a; echo A > a~; cp -b -V simple a~ a
+		 cd /tmp; rm -f a a~; : > a; echo A > a~; cp --b=simple a~ a
 		 would leave two zero-length files: a and a~.  */
+	      /* FIXME: but simply change e.g., the final a~ to `./a~'
+		 and the source will still be destroyed.  */
 	      if (STREQ (tmp_backup, src_path))
 		{
 		  const char *fmt;
-		  fmt = (move_mode
+		  fmt = (x->move_mode
 		 ? _("backing up %s would destroy source;  %s not moved")
 		 : _("backing up %s would destroy source;  %s not copied"));
 		  error (0, 0, fmt,
@@ -763,10 +1057,42 @@ copy_internal (const char *src_path, const char *dst_path,
       putchar ('\n');
     }
 
+  /* Associate the destination path with the source device and inode
+     so that if we encounter a matching dev/ino pair in the source tree
+     we can arrange to create a hard link between the corresponding names
+     in the destination tree.
+
+     Sometimes, when preserving links, we have to record dev/ino even
+     though st_nlink == 1:
+     - when using -H and processing a command line argument;
+	that command line argument could be a symlink pointing to another
+	command line argument.  With `cp -H --preserve=link', we hard-link
+	those two destination files.
+     - likewise for -L except that it applies to all files, not just
+	command line arguments.
+
+     Also record directory dev/ino when using --recursive.  We'll use that
+     info to detect this problem: cp -R dir dir.  FIXME-maybe: ideally,
+     directory info would be recorded in a separate hash table, since
+     such entries are useful only while a single command line hierarchy
+     is being copied -- so that separate table could be cleared between
+     command line args.  Using the same hash table to preserve hard
+     links means that it may not be cleared.  */
+
+  if ((x->preserve_links
+       && (1 < src_sb.st_nlink
+	   || (command_line_arg
+	       && x->dereference == DEREF_COMMAND_LINE_ARGUMENTS)
+	   || x->dereference == DEREF_ALWAYS))
+      || (x->recursive && S_ISDIR (src_type)))
+    {
+      earlier_file = remember_copied (dst_path, src_sb.st_ino, src_sb.st_dev);
+    }
+
   /* Did we copy this inode somewhere else (in this command line argument)
      and therefore this is a second hard link to the inode?  */
 
-  if (x->dereference == DEREF_NEVER && src_sb.st_nlink > 1 && earlier_file)
+  if (earlier_file)
     {
       /* Avoid damaging the destination filesystem by refusing to preserve
 	 hard-linked directories (which are found at least in Netapp snapshot
@@ -801,6 +1127,8 @@ copy_internal (const char *src_path, const char *dst_path,
       return 0;
     }
 
+  /* Note that this is testing the local variable move_mode, not
+     the x->move_mode member.  */
   if (move_mode)
     {
       if (rename (src_path, dst_path) == 0)
@@ -809,6 +1137,19 @@ copy_internal (const char *src_path, const char *dst_path,
 	    printf ("%s -> %s\n", quote_n (0, src_path), quote_n (1, dst_path));
 	  if (rename_succeeded)
 	    *rename_succeeded = 1;
+
+	  if (command_line_arg)
+	    {
+	      /* Record destination dev/ino/filename, so that if we are asked
+		 to overwrite that file again, we can detect it and fail.  */
+	      /* It's fine to use the _source_ stat buffer (src_sb) to get the
+	         _destination_ dev/ino, since the rename above can't have
+		 changed those, and `mv' always uses lstat.
+		 We could limit it further by operating
+		 only on non-directories.  */
+	      record_file (x->dest_info, dst_path, &src_sb);
+	    }
+
 	  return 0;
 	}
 
@@ -836,6 +1177,11 @@ copy_internal (const char *src_path, const char *dst_path,
 	  error (0, 0, _("cannot move %s to a subdirectory of itself, %s"),
 		 quote_n (0, top_level_src_path),
 		 quote_n (1, top_level_dst_path));
+
+	  /* Note that there is no need to call forget_created here,
+	     (compare with the other calls in this file) since the
+	     destination directory didn't exist before.  */
+
 	  *copy_into_self = 1;
 	  /* FIXME-cleanup: Don't return zero here; adjust mv.c accordingly.
 	     The only caller that uses this code (mv.c) ends up setting its
@@ -843,20 +1189,49 @@ copy_internal (const char *src_path, const char *dst_path,
 	  return 0;
 	}
 
-      /* Ignore other types of failure (e.g. EXDEV), since the following
-	 code will try to perform a copy, then remove.  */
-
-      /* Save this value of errno to use in case the unlink fails.  */
-      rename_errno = errno;
+      /* WARNING: there probably exist systems for which an inter-device
+	 rename fails with a value of errno not handled here.
+	 If/as those are reported, add them to the condition below.
+	 If this happens to you, please do the following and send the output
+	 to the bug-reporting address (e.g., in the output of cp --help):
+	   touch k; perl -e 'rename "k","/tmp/k" or print "$!(",$!+0,")\n"'
+	 where your current directory is on one partion and /tmp is the other.
+	 Also, please try to find the E* errno macro name corresponding to
+	 the diagnostic and parenthesized integer, and include that in your
+	 e-mail.  One way to do that is to run a command like this
+	   find /usr/include/. -type f \
+	     | xargs grep 'define.*\<E[A-Z]*\>.*\<18\>' /dev/null
+	 where you'd replace `18' with the integer in parentheses that
+	 was output from the perl one-liner above.
+	 If necessary, of course, change `/tmp' to some other directory.  */
+      if (errno != EXDEV)
+	{
+	  /* There are many ways this can happen due to a race condition.
+	     When something happens between the initial xstat and the
+	     subsequent rename, we can get many different types of errors.
+	     For example, if the destination is initially a non-directory
+	     or non-existent, but it is created as a directory, the rename
+	     fails.  If two `mv' commands try to rename the same file at
+	     about the same time, one will succeed and the other will fail.
+	     If the permissions on the directory containing the source or
+	     destination file are made too restrictive, the rename will
+	     fail.  Etc.  */
+	  error (0, errno,
+		 _("cannot move %s to %s"),
+		 quote_n (0, src_path), quote_n (1, dst_path));
+	  forget_created (src_sb.st_ino, src_sb.st_dev);
+	  return 1;
+	}
 
       /* The rename attempt has failed.  Remove any existing destination
 	 file so that a cross-device `mv' acts as if it were really using
 	 the rename syscall.  */
       if (unlink (dst_path) && errno != ENOENT)
 	{
-	  /* Use the value of errno from the failed rename.  */
-	  error (0, rename_errno, _("cannot move %s to %s"),
+	  error (0, errno,
+	     _("inter-device move failed: %s to %s; unable to remove target"),
 		 quote_n (0, src_path), quote_n (1, dst_path));
+	  forget_created (src_sb.st_ino, src_sb.st_dev);
 	  return 1;
 	}
 
@@ -864,6 +1239,13 @@ copy_internal (const char *src_path, const char *dst_path,
     }
 
   delayed_fail = 0;
+
+  /* In certain modes (cp's --symbolic-link), and for certain file types
+     (symlinks and hard links) it doesn't make sense to preserve metadata,
+     or it's possible to preserve only some of it.
+     In such cases, set this variable to zero.  */
+  preserve_metadata = 1;
+
   if (S_ISDIR (src_type))
     {
       struct dir_list *dir;
@@ -928,6 +1310,8 @@ copy_internal (const char *src_path, const char *dst_path,
 #ifdef S_ISLNK
   else if (x->symbolic_link)
     {
+      preserve_metadata = 0;
+
       if (*src_path != '/')
 	{
 	  /* Check that DST_PATH denotes a file in the current directory.  */
@@ -937,8 +1321,6 @@ copy_internal (const char *src_path, const char *dst_path,
 	  int in_current_dir;
 
 	  dst_parent = dir_name (dst_path);
-	  if (dst_parent == NULL)
-	    xalloc_die ();
 
 	  in_current_dir = (STREQ (".", dst_parent)
 			    /* If either stat call fails, it's ok not to report
@@ -963,18 +1345,16 @@ copy_internal (const char *src_path, const char *dst_path,
 		 quote_n (0, dst_path), quote_n (1, src_path));
 	  goto un_backup;
 	}
-
-      return 0;
     }
 #endif
   else if (x->hard_link)
     {
+      preserve_metadata = 0;
       if (link (src_path, dst_path))
 	{
 	  error (0, errno, _("cannot create link %s"), quote (dst_path));
 	  goto un_backup;
 	}
-      return 0;
     }
   else if (S_ISREG (src_type)
 	   || (x->copy_as_regular && !S_ISDIR (src_type)
@@ -988,7 +1368,7 @@ copy_internal (const char *src_path, const char *dst_path,
 	 used as the 3rd argument in the open call, but that's not consistent
 	 with historical practice.  */
       if (copy_reg (src_path, dst_path, x,
-		    get_dest_mode (x, src_mode), &new_dst))
+		    get_dest_mode (x, src_mode), &new_dst, &src_sb))
 	goto un_backup;
     }
   else
@@ -1020,35 +1400,31 @@ copy_internal (const char *src_path, const char *dst_path,
 #ifdef S_ISLNK
   if (S_ISLNK (src_type))
     {
-      char *link_val;
-      int link_size;
-
-      link_val = (char *) alloca (PATH_MAX + 2);
-      link_size = readlink (src_path, link_val, PATH_MAX + 1);
-      if (link_size < 0)
+      char *src_link_val = xreadlink (src_path);
+      if (src_link_val == NULL)
 	{
 	  error (0, errno, _("cannot read symbolic link %s"), quote (src_path));
 	  goto un_backup;
 	}
-      link_val[link_size] = '\0';
 
-      if (symlink (link_val, dst_path))
+      if (!symlink (src_link_val, dst_path))
+	free (src_link_val);
+      else
 	{
 	  int saved_errno = errno;
 	  int same_link = 0;
 	  if (x->update && !new_dst && S_ISLNK (dst_sb.st_mode))
 	    {
 	      /* See if the destination is already the desired symlink.  */
-	      char *dest_link_name = (char *) alloca (PATH_MAX + 2);
-	      int dest_link_len = readlink (dst_path, dest_link_name,
-					    PATH_MAX + 1);
-	      if (dest_link_len > 0)
-		{
-		  dest_link_name[dest_link_len] = '\0';
-		  if (STREQ (dest_link_name, link_val))
-		    same_link = 1;
-		}
+	      size_t src_link_len = strlen (src_link_val);
+	      char *dest_link_val = (char *) alloca (src_link_len + 1);
+	      int dest_link_len = readlink (dst_path, dest_link_val,
+					    src_link_len + 1);
+	      if ((size_t) dest_link_len == src_link_len
+		  && strncmp (dest_link_val, src_link_val, src_link_len) == 0)
+		same_link = 1;
 	    }
+	  free (src_link_val);
 
 	  if (! same_link)
 	    {
@@ -1058,14 +1434,18 @@ copy_internal (const char *src_path, const char *dst_path,
 	    }
 	}
 
-      if (x->preserve_owner_and_group)
+      /* There's no need to preserve timestamps or permissions.  */
+      preserve_metadata = 0;
+
+      if (x->preserve_ownership)
 	{
 	  /* Preserve the owner and group of the just-`copied'
 	     symbolic link, if possible.  */
 # if HAVE_LCHOWN
 	  if (DO_CHOWN (lchown, dst_path, src_sb.st_uid, src_sb.st_gid))
 	    {
-	      error (0, errno, _("preserving ownership for %s"), dst_path);
+	      error (0, errno, _("failed to preserve ownership for %s"),
+		     dst_path);
 	      goto un_backup;
 	    }
 # else
@@ -1075,8 +1455,6 @@ copy_internal (const char *src_path, const char *dst_path,
 	     preserving owner/group is a potential security problem.  */
 # endif
 	}
-
-      return 0;
     }
   else
 #endif
@@ -1085,6 +1463,12 @@ copy_internal (const char *src_path, const char *dst_path,
       goto un_backup;
     }
 
+  if (command_line_arg)
+    record_file (x->dest_info, dst_path, NULL);
+
+  if ( ! preserve_metadata)
+    return 0;
+
   /* POSIX says that `cp -p' must restore the following:
      - permission bits
      - setuid, setgid bits
@@ -1117,18 +1501,30 @@ copy_internal (const char *src_path, const char *dst_path,
     }
 
   /* Avoid calling chown if we know it's not necessary.  */
-  if (x->preserve_owner_and_group
+  if (x->preserve_ownership
       && (new_dst || !SAME_OWNER_AND_GROUP (src_sb, dst_sb)))
     {
       ran_chown = 1;
       if (DO_CHOWN (chown, dst_path, src_sb.st_uid, src_sb.st_gid))
 	{
-	  error (0, errno, _("preserving ownership for %s"), quote (dst_path));
+	  error (0, errno, _("failed to preserve ownership for %s"),
+		 quote (dst_path));
 	  if (x->require_preserve)
 	    return 1;
 	}
     }
 
+#if HAVE_STRUCT_STAT_ST_AUTHOR
+  /* Preserve the st_author field.  */
+  {
+    file_t file = getdport (dst_path);
+    if (file_chauthor (file, src_sb.st_author))
+      error (0, errno, _("failed to preserve authorship for %s"),
+	     quote (dst_path));
+    mach_port_deallocate (mach_task_self (), file);
+  }
+#endif
+
   /* Permissions of newly-created regular files were set upon `open' in
      copy_reg.  But don't return early if there were any special bits and
      we had to run chown, because the chown must have reset those bits.  */
@@ -1136,7 +1532,7 @@ copy_internal (const char *src_path, const char *dst_path,
       && !(ran_chown && (src_mode & ~S_IRWXUGO)))
     return delayed_fail;
 
-  if ((x->preserve_chmod_bits || new_dst)
+  if ((x->preserve_mode || new_dst)
       && (x->copy_as_regular || S_ISREG (src_type) || S_ISDIR (src_type)))
     {
       if (chmod (dst_path, get_dest_mode (x, src_mode)))
@@ -1150,6 +1546,13 @@ copy_internal (const char *src_path, const char *dst_path,
   return delayed_fail;
 
 un_backup:
+
+  /* We didn't create the destination.
+     Remove the entry associating the source dev/ino with the
+     destination file name, so we don't try to `preserve' a link
+     to a file we didn't create.  */
+  forget_created (src_sb.st_ino, src_sb.st_dev);
+
   if (dst_backup)
     {
       if (rename (dst_backup, dst_path))
@@ -1197,11 +1600,6 @@ copy (const char *src_path, const char *dst_path,
       int nonexistent_dst, const struct cp_options *options,
       int *copy_into_self, int *rename_succeeded)
 {
-  /* move_mode is set to the value from the `options' parameter for the
-     first copy_internal call.  For all subsequent calls (if any), it must
-     be zero.  */
-  int move_mode = options->move_mode;
-
   assert (valid_options (options));
 
   /* Record the file names: they're used in case of error, when copying
@@ -1210,10 +1608,10 @@ copy (const char *src_path, const char *dst_path,
      exceptional cases, but in this case, I don't see a way to derive the
      top level source and destination directory names where they're used.
      An alternative is to use COPY_INTO_SELF and print the diagnostic
-     from every caller -- but I don't wan't to do that.  */
+     from every caller -- but I don't want to do that.  */
   top_level_src_path = src_path;
   top_level_dst_path = dst_path;
 
   return copy_internal (src_path, dst_path, nonexistent_dst, 0, NULL,
-			options, move_mode, copy_into_self, rename_succeeded);
+			options, 1, copy_into_self, rename_succeeded);
 }
