@@ -1,5 +1,5 @@
 /* `ln' program to create links between files.
-   Copyright (C) 86, 89, 90, 91, 1995-2001 Free Software Foundation, Inc.
+   Copyright (C) 86, 89, 90, 91, 1995-2002 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,13 +29,14 @@
 #include "system.h"
 #include "same.h"
 #include "backupfile.h"
+#include "dirname.h"
 #include "error.h"
 #include "quote.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "ln"
 
-#define AUTHORS "Mike Parker and David MacKenzie"
+#define AUTHORS N_ ("Mike Parker and David MacKenzie")
 
 /* For long options that have no equivalent short option, use a
    non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
@@ -86,7 +87,6 @@ int symlink ();
 
 int isdir ();
 int yesno ();
-void strip_trailing_slashes ();
 
 /* The name by which the program was run, for error messages.  */
 char *program_name;
@@ -224,7 +224,7 @@ do_link (const char *source, const char *dest)
 	 equivalent: `ln -f k k' (with or without --backup) to get
 	 beyond this point, because the error message you'd get is
 	 misleading.  */
-      && (backup_type == none || !symlink)
+      && (backup_type == none || !symbolic_link)
       && (!symbolic_link || stat (source, &source_stats) == 0)
       && source_stats.st_dev == dest_stats.st_dev
       && source_stats.st_ino == dest_stats.st_ino
@@ -342,7 +342,7 @@ Usage: %s [OPTION]... TARGET [LINK_NAME]\n\
   or:  %s [OPTION]... --target-directory=DIRECTORY TARGET...\n\
 "),
 	      program_name, program_name, program_name);
-      printf (_("\
+      fputs (_("\
 Create a link to the specified TARGET with optional LINK_NAME.\n\
 If LINK_NAME is omitted, a link with the same basename as the TARGET is\n\
 created in the current directory.  When using the second form with more\n\
@@ -350,33 +350,44 @@ than one TARGET, the last argument must be a directory;  create links\n\
 in DIRECTORY to each TARGET.  Create hard links by default, symbolic\n\
 links with --symbolic.  When creating hard links, each TARGET must exist.\n\
 \n\
+"), stdout);
+      fputs (_("\
+Mandatory arguments to long options are mandatory for short options too.\n\
+"), stdout);
+      fputs (_("\
       --backup[=CONTROL]      make a backup of each existing destination file\n\
   -b                          like --backup but does not accept an argument\n\
   -d, -F, --directory         hard link directories (super-user only)\n\
   -f, --force                 remove existing destination files\n\
+"), stdout);
+      fputs (_("\
   -n, --no-dereference        treat destination that is a symlink to a\n\
                                 directory as if it were a normal file\n\
   -i, --interactive           prompt whether to remove destinations\n\
   -s, --symbolic              make symbolic links instead of hard links\n\
+"), stdout);
+      fputs (_("\
   -S, --suffix=SUFFIX         override the usual backup suffix\n\
       --target-directory=DIRECTORY  specify the DIRECTORY in which to create\n\
                                 the links\n\
   -v, --verbose               print name of each file before linking\n\
-      --help                  display this help and exit\n\
-      --version               output version information and exit\n\
+"), stdout);
+      fputs (HELP_OPTION_DESCRIPTION, stdout);
+      fputs (VERSION_OPTION_DESCRIPTION, stdout);
+      fputs (_("\
 \n\
-"));
-      printf (_("\
 The backup suffix is `~', unless set with --suffix or SIMPLE_BACKUP_SUFFIX.\n\
 The version control method may be selected via the --backup option or through\n\
 the VERSION_CONTROL environment variable.  Here are the values:\n\
 \n\
+"), stdout);
+      fputs (_("\
   none, off       never make backups (even if --backup is given)\n\
   numbered, t     make numbered backups\n\
   existing, nil   numbered if numbered backups exist, simple otherwise\n\
   simple, never   always make simple backups\n\
-"));
-      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+"), stdout);
+      printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
   exit (status);
 }
@@ -506,7 +517,7 @@ main (int argc, char **argv)
 
   if (target_directory_specified && !dest_is_dir)
     {
-      error (0, 0, _("specified target directory, %s is not a directory"),
+      error (0, 0, _("%s: specified target directory is not a directory"),
 	     quote (target_directory));
       usage (1);
     }
