@@ -1,5 +1,5 @@
 /* mkdir -- make directories
-   Copyright (C) 90, 1995-2000 Free Software Foundation, Inc.
+   Copyright (C) 90, 1995-2001 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@
 #include <sys/types.h>
 
 #include "system.h"
+#include "dirname.h"
 #include "error.h"
 #include "makepath.h"
 #include "modechange.h"
@@ -58,16 +59,21 @@ usage (int status)
   else
     {
       printf (_("Usage: %s [OPTION] DIRECTORY...\n"), program_name);
-      printf (_("\
+      fputs (_("\
 Create the DIRECTORY(ies), if they do not already exist.\n\
 \n\
+"), stdout);
+      fputs (_("\
+Mandatory arguments to long options are mandatory for short options too.\n\
+"), stdout);
+      fputs (_("\
   -m, --mode=MODE   set permission mode (as in chmod), not rwxrwxrwx - umask\n\
   -p, --parents     no error if existing, make parent directories as needed\n\
   -v, --verbose     print a message for each created directory\n\
-      --help        display this help and exit\n\
-      --version     output version information and exit\n\
-"));
-      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+"), stdout);
+      fputs (HELP_OPTION_DESCRIPTION, stdout);
+      fputs (VERSION_OPTION_DESCRIPTION, stdout);
+      printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
   exit (status);
 }
@@ -77,7 +83,7 @@ main (int argc, char **argv)
 {
   mode_t newmode;
   mode_t parent_mode;
-  const char *symbolic_mode = NULL;
+  const char *specified_mode = NULL;
   const char *verbose_fmt_string = NULL;
   int errors = 0;
   int optc;
@@ -101,7 +107,7 @@ main (int argc, char **argv)
 	  create_parents = 1;
 	  break;
 	case 'm':
-	  symbolic_mode = optarg;
+	  specified_mode = optarg;
 	  break;
 	case 'v': /* --verbose  */
 	  verbose_fmt_string = _("created directory %s");
@@ -119,13 +125,19 @@ main (int argc, char **argv)
       usage (1);
     }
 
-  newmode = S_IRWXUGO & ~ umask (0);
-  parent_mode = S_IWUSR | S_IXUSR | newmode;
-  if (symbolic_mode)
+  newmode = S_IRWXUGO;
+  {
+    mode_t umask_value = umask (0);
+    umask (umask_value);		/* Restore the old value. */
+    parent_mode = (newmode & (~ umask_value)) | S_IWUSR | S_IXUSR;
+  }
+
+  if (specified_mode)
     {
-      struct mode_change *change = mode_compile (symbolic_mode, 0);
+      struct mode_change *change = mode_compile (specified_mode, 0);
+      newmode &= ~ umask (0);
       if (change == MODE_INVALID)
-	error (1, 0, _("invalid mode %s"), quote (symbolic_mode));
+	error (1, 0, _("invalid mode %s"), quote (specified_mode));
       else if (change == MODE_MEMORY_EXHAUSTED)
 	xalloc_die ();
       newmode = mode_adjust (newmode, change);
@@ -134,33 +146,56 @@ main (int argc, char **argv)
   for (; optind < argc; ++optind)
     {
       int fail = 0;
+
       if (create_parents)
 	{
-	  fail = make_path (argv[optind], newmode, parent_mode,
+	  char *parents = dir_name (argv[optind]);
+	  fail = make_path (parents, parent_mode, parent_mode,
 			    -1, -1, 1, verbose_fmt_string);
+	  free (parents);
 	}
-      else
+
+      if (fail == 0)
 	{
-	  fail = mkdir (argv[optind], newmode);
+	  const char *dir = argv[optind];
+	  int dir_created;
+	  fail = make_dir (dir, dir, newmode, &dir_created);
 	  if (fail)
-	    error (0, errno, _("cannot create directory %s"),
-		   quote (argv[optind]));
+	    {
+	      /* make_dir already gave a diagnostic.  */
+	    }
+	  else if (!create_parents && !dir_created)
+	    {
+	      /* make_dir `succeeds' when DIR already exists.
+		 In that case, mkdir must fail, unless --parents (-p)
+		 was specified.  */
+	      error (0, EEXIST, _("cannot create directory %s"),
+		     quote (dir));
+	      fail = 1;
+	    }
 	  else if (verbose_fmt_string)
-	    error (0, 0, verbose_fmt_string, quote (argv[optind]));
+	    error (0, 0, verbose_fmt_string, quote (dir));
 
 	  /* mkdir(2) is required to honor only the file permission bits.
 	     In particular, it needn't do anything about `special' bits,
-	     so if any were set in newmode, apply them with chmod.  */
-	  if (fail == 0 && (newmode & ~S_IRWXUGO))
+	     so if any were set in newmode, apply them with chmod.
+	     This extra step is necessary in some cases when the containing
+	     directory has a default ACL.  */
+
+	  /* Set the permissions only if this directory has just
+	     been created.  */
+
+	  if (fail == 0 && specified_mode && dir_created)
 	    {
-	      fail = chmod (argv[optind], newmode);
+	      fail = chmod (dir, newmode);
 	      if (fail)
 		error (0, errno, _("cannot set permissions of directory %s"),
-		       quote (argv[optind]));
+		       quote (dir));
 	    }
 	}
 
-      errors |= fail;
+      if (fail)
+	errors = 1;
     }
 
   exit (errors);
