@@ -1,5 +1,5 @@
 /* mknod -- make special files
-   Copyright (C) 90, 91, 1995-2000 Free Software Foundation, Inc.
+   Copyright (C) 90, 91, 1995-2001 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -63,20 +63,27 @@ usage (int status)
   else
     {
       printf (_("Usage: %s [OPTION]... NAME TYPE [MAJOR MINOR]\n"), program_name);
-      printf (_("\
+      fputs (_("\
 Create the special file NAME of the given TYPE.\n\
 \n\
+"), stdout);
+      fputs (_("\
+Mandatory arguments to long options are mandatory for short options too.\n\
+"), stdout);
+      fputs (_("\
   -m, --mode=MODE   set permission mode (as in chmod), not a=rw - umask\n\
-      --help        display this help and exit\n\
-      --version     output version information and exit\n\
+"), stdout);
+      fputs (HELP_OPTION_DESCRIPTION, stdout);
+      fputs (VERSION_OPTION_DESCRIPTION, stdout);
+      fputs (_("\
 \n\
 MAJOR MINOR are forbidden for TYPE p, mandatory otherwise.  TYPE may be:\n\
 \n\
   b      create a block (buffered) special file\n\
   c, u   create a character (unbuffered) special file\n\
   p      create a FIFO\n\
-"));
-      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+"), stdout);
+      printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
   exit (status);
 }
@@ -88,9 +95,7 @@ main (int argc, char **argv)
   struct mode_change *change;
   const char *specified_mode;
   int optc;
-  int i_major, i_minor;
-  long int tmp_major, tmp_minor;
-  char *s;
+  mode_t node_type;
 
   program_name = argv[0];
   setlocale (LC_ALL, "");
@@ -138,7 +143,7 @@ main (int argc, char **argv)
 	msg = _("too many arguments");
       else
 	msg = _("wrong number of arguments");
-      error (0, 0, msg);
+      error (0, 0, "%s", msg);
       usage (1);
     }
 
@@ -151,57 +156,51 @@ main (int argc, char **argv)
 #ifndef S_IFBLK
       error (4, 0, _("block special files not supported"));
 #else
-      if (argc - optind != 4)
-	{
-	  error (0, 0, _("\
-when creating block special files, major and minor device\n\
-numbers must be specified"));
-	  usage (1);
-	}
-
-      s = argv[optind + 2];
-      if (xstrtol (s, NULL, 0, &tmp_major, NULL) != LONGINT_OK)
-	error (1, 0, _("invalid major device number %s"), quote (s));
-
-      s = argv[optind + 3];
-      if (xstrtol (s, NULL, 0, &tmp_minor, NULL) != LONGINT_OK)
-	error (1, 0, _("invalid minor device number %s"), quote (s));
-
-      i_major = (int) tmp_major;
-      i_minor = (int) tmp_minor;
-
-      if (mknod (argv[optind], newmode | S_IFBLK, makedev (i_major, i_minor)))
-	error (1, errno, "%s", quote (argv[optind]));
+      node_type = S_IFBLK;
 #endif
-      break;
+      goto block_or_character;
 
     case 'c':			/* `character' */
     case 'u':			/* `unbuffered' */
 #ifndef S_IFCHR
       error (4, 0, _("character special files not supported"));
 #else
+      node_type = S_IFCHR;
+#endif
+      goto block_or_character;
+
+    block_or_character:
       if (argc - optind != 4)
 	{
 	  error (0, 0, _("\
-when creating character special files, major and minor device\n\
+when creating special files, major and minor device\n\
 numbers must be specified"));
 	  usage (1);
 	}
 
-      s = argv[optind + 2];
-      if (xstrtol (s, NULL, 0, &tmp_major, NULL) != LONGINT_OK)
-	error (1, 0, _("invalid major device number %s"), quote (s));
+      {
+	char const *s_major = argv[optind + 2];
+	char const *s_minor = argv[optind + 3];
+	uintmax_t i_major, i_minor;
+	dev_t device;
 
-      s = argv[optind + 3];
-      if (xstrtol (s, NULL, 0, &tmp_minor, NULL) != LONGINT_OK)
-	error (1, 0, _("invalid minor device number %s"), quote (s));
+	if (xstrtoumax (s_major, NULL, 0, &i_major, NULL) != LONGINT_OK
+	    || i_major != (major_t) i_major)
+	  error (1, 0, _("invalid major device number %s"), quote (s_major));
 
-      i_major = (int) tmp_major;
-      i_minor = (int) tmp_minor;
+	if (xstrtoumax (s_minor, NULL, 0, &i_minor, NULL) != LONGINT_OK
+	    || i_minor != (minor_t) i_minor)
+	  error (1, 0, _("invalid minor device number %s"), quote (s_minor));
 
-      if (mknod (argv[optind], newmode | S_IFCHR, makedev (i_major, i_minor)))
-	error (1, errno, "%s", quote (argv[optind]));
+	device = makedev (i_major, i_minor);
+#ifdef NODEV
+	if (device == NODEV)
+	  error (1, 0, _("invalid device %s %s"), s_major, s_minor);
 #endif
+
+	if (mknod (argv[optind], newmode | node_type, device) != 0)
+	  error (1, errno, "%s", quote (argv[optind]));
+      }
       break;
 
     case 'p':			/* `pipe' */
@@ -220,6 +219,7 @@ major and minor device numbers may not be specified for fifo files"));
       break;
 
     default:
+      error (0, 0, "invalid device type %s", quote (argv[optind + 1]));
       usage (1);
     }
 
@@ -230,7 +230,7 @@ major and minor device numbers may not be specified for fifo files"));
   if (specified_mode)
     {
       if (chmod (argv[optind], newmode))
-        error (0, errno, _("cannot set permissions of `%s'"),
+        error (0, errno, _("cannot set permissions of %s"),
                quote (argv[optind]));
     }
 
